# -*- coding: utf-8 -*-

"""
    Unit tests for the divergence module of the SmartSVM package.
"""

from __future__ import division

import unittest
import numpy as np

from smartsvm import divergence


class MstTestCase(unittest.TestCase):
    def test_merge_and_label(self):
        """ DIVERGENCE: Test merging and labelling of datasets """
        X = np.array([[0, 1, 2], [3, 2, 1]])
        Y = np.array([[9, 8, 7], [6, 5, 4], [0, 1, 0]])
        data, labels = divergence.merge_and_label(X, Y)
        self.assertTrue(
            np.array_equal(
                data,
                np.array(
                    [[0, 1, 2], [3, 2, 1], [9, 8, 7], [6, 5, 4], [0, 1, 0]]
                ),
            )
        )
        self.assertTrue(
            np.array_equal(labels, np.array([-1.0, -1.0, 1.0, 1.0, 1.0]))
        )

    def test_divergence_correction_1(self):
        """ DIVERGENCE: Testing divergence measure on small dataset. """
        X = np.array([[-1, -1], [-2, 4], [-5, -5]])
        Y = np.array([[4, 1], [3, -3], [5, 4]])
        div = divergence.divergence(X, Y)
        self.assertAlmostEqual(div, 1.0 / 3.0)

    def test_divergence_no_correction_1(self):
        """ DIVERGENCE: Testing divergence measure on small dataset. """
        X = np.array([[-1, -1], [-2, 4], [-5, -5]])
        Y = np.array([[4, 1], [3, -3], [5, 4]])
        div = divergence.divergence(X, Y, bias_correction=False)
        self.assertAlmostEqual(div, 1.0 / 3.0)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

from __future__ import division

import networkx as nx
import numpy as np

from sklearn.svm import LinearSVC
from sklearn.utils import check_X_y
from sklearn.utils.multiclass import check_classification_targets

from .base import HierarchicalClassifier
from .cut import graph_cut
from .error_estimate import compute_error_graph


class SmartSVM(HierarchicalClassifier):
    """SmartSVM classifier for multiclass classification

    This is the SmartSVM classifier. This classifier splits the classes into a 
    hierarchy based on the weighted complete graph of pairwise estimates of the 
    Bayes Error Rate between classes.

    Note that this class is used in a binary tree. It therefore has children 
    which are also elements of the :class:`.SmartSVM` class.

    This class inherits from the :class:`.HierarchicalClassifier` class, more 
    documentation on the basic methods of the class can be found there.

    Parameters
    ----------

    binary_clf : classifier
        The type of the binary classifier to be used for each binary 
        subproblem. This will typically be a scikit-learn classifier such as 
        LinearSVC, DecisionTreeClassifier, etc.

    clf_params : dict
        Parameters to pass on to the constructor of the ``binary_clf`` type 
        classifier. It must be a dict with a mapping from parameter to value.

    n_jobs : int
        Number of parallel jobs to use where applicable.

    graph : networkx.Graph
        Complete weighted graph with the pairwise Bayes error estimates. If it 
        is not supplied, it will be computed when the :meth:`.fit` method is 
        called.  This parameter exists for situations when the graph is 
        precomputed.

    cut_algorithm : str
        The algorithm to use for the graph cut. Several algorithms are 
        currently implemented, see :func:`.cut.graph_cut` for more details.

    normalize_error : bool
        Whether or not to normalize the Bayes error estimates with the 
        empirical estimate of the prior probability. 


    """

    def __init__(
        self,
        binary_clf=LinearSVC,
        clf_params=None,
        n_jobs=1,
        graph=None,
        cut_algorithm="stoer_wagner",
        normalize_error=True,
    ):
        super(SmartSVM, self).__init__(
            binary_clf=binary_clf, clf_params=clf_params, n_jobs=n_jobs
        )
        self.graph = graph
        self.cut_algorithm = cut_algorithm
        self.normalize_error = normalize_error

    @property
    def elements(self):
        """ The labels of this node in the hierarchy """
        if not isinstance(self.graph, nx.Graph):
            raise ValueError("graph should be a networkx Graph object.")
        return list(self.graph.nodes())

    def fit(self, X, y):
        """Fit the SmartSVM classifier

        This method fits the SmartSVM classifier on the given data. If the 
        ``graph`` attribute of the class is not defined, it will be computed 
        with :func:`.error_estimate.compute_error_graph`. If this node in the 
        hierarchy can not be split further, no classifier will be fit.  
        Otherwise, the node will be split to form a binary classification 
        problem. Next, the binary classifier will be trained on this problem.  
        Finally, the left and right children of the classifier will be trained 
        as a recursive step.

        **Important:** This method constructs the ``graph`` attribute of the 
        class if it does not exist yet. However, if the graph does exist, it is 
        not recomputed. This means that a problem can occur if you construct an 
        instance of the :class:`.SmartSVM` classifier, fit it, then fit it 
        again with different data. Namely, the graph for the first data may not 
        be appropriate for the second dataset. The solution is however simple: 
        when fitting with a different dataset, reset the graph using 
        ``clf.graph = None``.

        Parameters
        ----------
        X : numpy.ndarray, accept csr sparse
            Training data, feature matrix
        y : numpy.ndarray
            Training data, label vector

        Returns
        -------
        self : object
            Returns self.

        """

        check_classification_targets(y)
        self.classes_ = np.unique(y)

        if not self.graph:
            X, y = check_X_y(
                X, y, accept_sparse="csr", dtype=np.float64, order="C"
            )
            self.graph = compute_error_graph(
                X, y, n_jobs=self.n_jobs, normalize=self.normalize_error
            )

        if not self.is_splitable:
            return

        self._split()
        self._fit_binary(X, y)

        self.negative_child_.fit(X, y)
        self.positive_child_.fit(X, y)

        return self

    def _split(self):
        """Split a leaf node in the binary tree if possible

        This method splits the current node into two leaf nodes of the binary 
        tree if possible. This is done by cutting the weighted error graph and 
        initializing the children using the obtained subgraphs.

        """
        if not self.is_splitable:
            return
        G_neg, G_pos = graph_cut(self.graph, algorithm=self.cut_algorithm)

        self.negative_ = set(G_neg.nodes())
        self.positive_ = set(G_pos.nodes())

        self.negative_child_ = SmartSVM(
            cut_algorithm=self.cut_algorithm,
            binary_clf=self.binary_clf,
            clf_params=self.clf_params,
            n_jobs=self.n_jobs,
            graph=G_neg,
            normalize_error=self.normalize_error,
        )

        self.positive_child_ = SmartSVM(
            cut_algorithm=self.cut_algorithm,
            binary_clf=self.binary_clf,
            clf_params=self.clf_params,
            n_jobs=self.n_jobs,
            graph=G_pos,
            normalize_error=self.normalize_error,
        )

    def __repr__(self):
        name = self.__class__.__name__
        if self.is_splitted:
            return "%s(negative=%r, positive=%r)" % (
                name,
                self.negative_,
                self.positive_,
            )
        elif self.graph is None:
            return "%s()" % (name)
        return "%s(%r)" % (name, set(self.elements))



"""
    Unit tests for the error estimate module of the SmartSVM package.
"""

from __future__ import division

import math
import unittest

import numpy as np
import networkx as nx

from smartsvm import error_estimate


class ErrorEstimateTestCase(unittest.TestCase):
    def test_worker_1(self):
        """ ERROR_ESTIMATE: Test worker """
        # This uses the same dataset in the test_divergence code, so we know
        # that the divergence is 1/3, so we can calculate what the error
        # estimate should be.
        X = np.array([[-1, -1], [-2, 4], [-5, -5], [4, 1], [3, -3], [5, 4]])
        y = np.array([1, 1, 1, 2, 2, 2])
        i = 1
        j = 2
        normalize = False
        Xyijn = (X, y, i, j, normalize)
        exp = (i, j, 1 / 2 - 1 / 4 * math.sqrt(1 / 3) - 1 / 4 * 1 / 3)
        res = error_estimate._worker(Xyijn)
        self.assertEqual(exp[0], res[0])
        self.assertEqual(exp[1], res[1])
        self.assertAlmostEqual(exp[2], res[2])

    def test_compute_error_graph(self):
        """ ERROR ESTIMATE: Test build error graph """
        # Test using a very simple dataset for which all pairwise errors are
        # the same, equal to 1/2 - 1/4*sqrt(1/3) - 1/4*1/3
        X = np.array(
            [
                [-0.05, 3],
                [0.05, 3],
                [0, 3.00],
                [3, 0.05],
                [3, -0.05],
                [3.00, 0],
                [-0.05, -3],
                [0.05, -3],
                [0, -3.00],
                [-3, 0.05],
                [-3, -0.05],
                [-3.00, 0],
            ]
        )
        y = np.array([1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4])
        # divide by 0.5 since we use the normalized error
        exp_weight = (
            1.0 / 2.0
            - 1.0 / 4.0 * math.sqrt(1.0 / 3.0)
            - 1.0 / 4.0 * 1.0 / 3.0
        ) / 0.5
        exp_G = nx.Graph()
        exp_G.add_edge(1, 2, weight=exp_weight)
        exp_G.add_edge(1, 3, weight=exp_weight)
        exp_G.add_edge(1, 4, weight=exp_weight)
        exp_G.add_edge(2, 3, weight=exp_weight)
        exp_G.add_edge(2, 4, weight=exp_weight)
        exp_G.add_edge(3, 4, weight=exp_weight)

        res_G_1 = error_estimate.compute_error_graph(X, y, n_jobs=1)
        em = nx.algorithms.isomorphism.numerical_edge_match("weight", 1)
        self.assertTrue(nx.is_isomorphic(res_G_1, exp_G, edge_match=em))
        self.assertEqual(sorted(res_G_1.nodes()), sorted(exp_G.nodes()))

        res_G_2 = error_estimate.compute_error_graph(X, y, n_jobs=2)
        em = nx.algorithms.isomorphism.numerical_edge_match("weight", 1)
        self.assertTrue(nx.is_isomorphic(res_G_2, exp_G, edge_match=em))
        self.assertEqual(sorted(res_G_2.nodes()), sorted(exp_G.nodes()))

    def test_compute_ovr_error_normalize(self):
        """ ERROR ESTIMATE: Test build error graph OvR (normalized)"""
        # Test using a very simple dataset. Error rates are checked using
        # Matlab.
        X = np.array(
            [
                [-0.05, 3],
                [0.05, 3],
                [0.00, 3],
                [3.1, 0.05],
                [3.1, -0.05],
                [3.1, 0.00],
                [-0.05, -3.3],
                [0.05, -3.3],
                [0.00, -3.3],
                [-3.7, 0.05],
                [-3.7, -0.05],
                [-3.7, 0.00],
            ]
        )
        y = np.array([1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4])
        N = X.shape[0]

        v1 = (
            1.0 / 2.0
            - 1.0 / 4.0 * math.sqrt(0.381966011250105)
            - 1.0 / 4.0 * 0.381966011250105
        ) / min(3 / N, (N - 3) / N)
        v2 = (
            1.0 / 2.0
            - 1.0 / 4.0 * math.sqrt(0.388888888888889)
            - 1.0 / 4.0 * 0.388888888888889
        ) / min(3 / N, (N - 3) / N)
        v3 = (
            1.0 / 2.0
            - 1.0 / 4.0 * math.sqrt(0.833333333333333)
            - 1.0 / 4.0 * 0.833333333333333
        ) / min(3 / N, (N - 3) / N)
        v4 = (
            1.0 / 2.0
            - 1.0 / 4.0 * math.sqrt(0.777777777777778)
            - 1.0 / 4.0 * 0.777777777777778
        ) / min(3 / N, (N - 3) / N)

        exp = {1: v1, 2: v2, 3: v3, 4: v4}

        res_1 = error_estimate.compute_ovr_error(X, y)
        res_2 = error_estimate.compute_ovr_error(X, y)
        for i in range(1, 5):
            self.assertAlmostEqual(exp[i], res_1[i])
            self.assertAlmostEqual(exp[i], res_2[i])

    def test_compute_ovr_error_no_normalize(self):
        """ ERROR ESTIMATE: Test build error graph OvR (not normalized)"""
        # Test using a very simple dataset. Error rates are checked using
        # Matlab.
        X = np.array(
            [
                [-0.05, 3],
                [0.05, 3],
                [0.00, 3],
                [3.1, 0.05],
                [3.1, -0.05],
                [3.1, 0.00],
                [-0.05, -3.3],
                [0.05, -3.3],
                [0.00, -3.3],
                [-3.7, 0.05],
                [-3.7, -0.05],
                [-3.7, 0.00],
            ]
        )
        y = np.array([1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4])

        exp = {
            1: (
                1.0 / 2.0
                - 1.0 / 4.0 * math.sqrt(0.381966011250105)
                - 1.0 / 4.0 * 0.381966011250105
            ),
            2: (
                1.0 / 2.0
                - 1.0 / 4.0 * math.sqrt(0.388888888888889)
                - 1.0 / 4.0 * 0.388888888888889
            ),
            3: (
                1.0 / 2.0
                - 1.0 / 4.0 * math.sqrt(0.833333333333333)
                - 1.0 / 4.0 * 0.833333333333333
            ),
            4: (
                1.0 / 2.0
                - 1.0 / 4.0 * math.sqrt(0.777777777777778)
                - 1.0 / 4.0 * 0.777777777777778
            ),
        }

        res_1 = error_estimate.compute_ovr_error(X, y, normalize=False)
        res_2 = error_estimate.compute_ovr_error(X, y, normalize=False)
        for i in range(1, 5):
            self.assertAlmostEqual(exp[i], res_1[i])
            self.assertAlmostEqual(exp[i], res_2[i])

    def test_hp_estimate_1(self):
        """ ERROR ESTIMATE: Test error estimate even unnormalized """
        X = np.array([[-1, -1], [-2, 4], [-5, -5]])
        Y = np.array([[4, 1], [3, -3], [5, 4]])

        exp = 5.0 / 12.0 - 1.0 / (4.0 * math.sqrt(3.0))
        res = error_estimate.hp_estimate(X, Y, normalize=False)
        self.assertEqual(res, exp)

    def test_hp_estimate_2(self):
        """ ERROR ESTIMATE: Test error estimate uneven unnormalized """
        X = np.array([[-1, -1], [-2, 4], [-5, -5]])
        Y = np.array([[4, 1], [3, -3], [5, 4], [3, 3]])

        exp = 0.22919372518014852
        res = error_estimate.hp_estimate(X, Y, normalize=False)
        self.assertEqual(res, exp)

    def test_hp_estimate_3(self):
        """ ERROR ESTIMATE: Test error estimate even normalized """
        X = np.array([[-1, -1], [-2, 4], [-5, -5]])
        Y = np.array([[4, 1], [3, -3], [5, 4]])

        exp = 5.0 / 12.0 - 1.0 / (4.0 * math.sqrt(3.0))
        exp = exp * 2.0
        res = error_estimate.hp_estimate(X, Y, normalize=True)
        self.assertEqual(res, exp)

    def test_hp_estimate_4(self):
        """ ERROR ESTIMATE: Test error estimate uneven normalized """
        X = np.array([[-1, -1], [-2, 4], [-5, -5]])
        Y = np.array([[4, 1], [3, -3], [5, 4], [3, 3]])

        exp = 0.22919372518014852
        exp = exp / (3.0 / 7.0)
        res = error_estimate.hp_estimate(X, Y, normalize=True)
        self.assertEqual(res, exp)


if __name__ == "__main__":
    unittest.main()



"""
Orthogonal MST Counter

Compute the first `k` orthogonal minimal spanning trees and return the number 
of edges that come from different samples.

"""

import cython

import numpy as np
cimport numpy as np

cdef extern void c_ortho_mst_count(double *weights, long *labels, long nTrees, 
        long nObs, long *count_ret, long *n_tree_ret)

@cython.boundscheck(False)
@cython.wraparound(False)
def ortho_mst_count(np.ndarray[double, ndim=2, mode="c"] weights not None,
        np.ndarray[long, ndim=1, mode="c"] labels not None, long nTrees):
    """
    ortho_mst_count(weights, labels, nTrees)

    Takes the weight matrix and constructs nTree minimal spanning trees. For 
    each minimal spanning tree the number of edges that connect points with 
    different labels is calculated. The total number of edges for which this 
    occurs is returned. For the MST algorithm the algorithm of Whitney is used 
    (see the C file).

    param: weights -- a 2-d symmetric numpy array of np.float64.
    param: labels -- a 1-d numpy array of np.intc
    param: nTrees -- the number of minimal spanning trees to construct

    """

    cdef long nObs
    cdef long count
    cdef long actual_trees

    nObs = weights.shape[0]

    c_ortho_mst_count(&weights[0, 0], &labels[0], nTrees, nObs, &count, 
            &actual_trees)

    return count, actual_trees



# -*- coding: utf-8 -*-

"""Code for computing the number of cross connections in an MST.

The functions in this module are used to compute the number of 
cross-connections between data points of different classes in the Euclidean 
Minimal Spanning Tree. It is the interface between the Cython code talking to 
the C implementations, and the higher level functions that compute the 
Henze-Penrose divergence.

"""

from __future__ import division

from scipy.spatial.distance import pdist, squareform

from .ortho_mst import ortho_mst_count
from .multiclass_mst_count import multiclass_mst_count


def binary_cross_connections(X, y, nTrees=3):
    """Compute the number of cross connections for two classes

    This function computes the average number of non-trivial cross-connections 
    in the orthogonal MSTs between points from different classes. For each MST 
    the number of times a point from one class is connected to a point from a 
    different class is recorded. This is reduced by 1 to adjust for the trivial 
    connection that will always exist, even in the case of large separation.  
    The result is averaged for each of the orthogonal MSTs.

    A warning can occur when the requested number of orthogonal MSTs can't be 
    computed on the data. This happens when there are either too few datapoints 
    to construct this many MSTs, or in the extreme case where each edge to a 
    data point is used in previous MSTs.

    Parameters
    ----------
    X : numpy.ndarray
        Data matrix

    y : np.ndarray
        Class labels, assumed to be only +1 and -1

    nTrees : int
        The number of orthogonal MSTs to compute

    Returns
    -------
    C : float
        The (average) number of non-trivial connections between instances of 
        different classes in the MSTs

    """

    # Calculate the Euclidean distance matrix
    weights = squareform(pdist(X, "euclidean"))

    # Number of edges connecting nodes from different classes
    S, actual_trees = ortho_mst_count(weights, y, nTrees)
    if actual_trees < nTrees:
        print(
            "Warning: instead of the requested %i orthogonal MSTs, %i "
            "orthogonal MSTs were used. Usually this is nothing to worry "
            "about." % (nTrees, actual_trees)
        )
    nTrees = actual_trees

    # Average this number and correct it slightly
    C = (S - nTrees) / nTrees

    return C


def ovr_cross_connections(X, y, nClass, nTrees=3):
    """Compute the One-vs-Rest cross-connection counts

    This function computes the cross-connection counts in the One-vs-Rest 
    setting by constructing orthogonal MSTs on the entire dataset, and 
    collecting cross-connection counts for each class.

    Parameters
    ----------

    X : numpy.ndarray
        Numpy array of the data matrix

    y : numpy.ndarray
        Numpy array of the class labels, assumed to be in 0..nClass-1

    nClass : int
        The number of classes in y for the full problem. This needs to be 
        supplied in case y is a subset of the full dataset where not all 
        classes are present. This ensures that the outcome matrix has the 
        appropriate size.

    nTrees : int
        The number of orthogonal MSTs to compute

    Returns
    -------
    Cs : numpy.ndarray
        Vector of class cross-connection counts, ordered by class label (nClass 
        x nClass)

    """
    full_set = set(range(nClass))
    if not set(y).issubset(full_set):
        raise ValueError("Class labels should be in {0, ..., nClass-1}")

    # Calculate the Euclidean distance matrix
    weights = squareform(pdist(X, "euclidean"))

    # Compute the cross-connection counts for each class
    # (matrix of nClass x nClass)
    C, actual_trees = multiclass_mst_count(weights, y, nClass, nTrees)

    # Throw a warning when not enough MSTs could be computed
    if actual_trees < nTrees:
        print(
            "Warning: instead of the requested %i orthogonal MSTs, %i "
            "orthogonal MSTs were used. Usually this is nothing to worry "
            "about." % (nTrees, actual_trees)
        )
    nTrees = actual_trees

    # Sum the matrix to obtain the class cross-connection counts and correct
    # for trivial connection
    sums = C.sum(0)
    sums[sums > 0] -= nTrees

    # Average over the number of MSTs computed
    Cs = sums / nTrees
    return Cs



"""
    Unit tests for the ortho_mst C extension of the SmartSVM package.
"""

from __future__ import division

import unittest
import numpy as np

from smartsvm import ortho_mst


class MstTestCase(unittest.TestCase):
    def test_empty(self):
        """ ORTHO_MST: Empty matrices should be okay. """
        weights = np.ones((0, 2), dtype=np.float_)
        labels = np.array((0,), dtype=np.int_)
        count, _ = ortho_mst.ortho_mst_count(weights, labels, 3)
        self.assertEqual(count, 0)

    def test_orthogonal_mst_1(self):
        """ ORTHO_MST: Making 3 orthogonal MST returns all edges"""
        # We need at least six nodes to make 3 orthogonal MSTs
        weights = np.array(
            [
                [0, 1, 3, 2, 2, 3],
                [1, 0, 2, 2, 1, 3],
                [3, 2, 0, 1, 3, 2],
                [2, 2, 1, 0, 1, 3],
                [2, 1, 3, 1, 0, 1],
                [3, 3, 2, 3, 1, 0],
            ],
            dtype=np.float_,
        )
        labels = np.array([-1, -1, -1, 1, 1, 1], dtype=np.int_)

        count, trees = ortho_mst.ortho_mst_count(weights, labels, 3)
        self.assertEqual(count, 9)
        self.assertEqual(trees, 3)

    def test_orthogonal_mst_2(self):
        """ ORTHO_MST: Making 3 orthogonal MST returns all edges"""
        # We need at least six nodes to make 3 orthogonal MSTs
        weights = np.array(
            [
                [0, 1, 3, 2, 2, 3],
                [1, 0, 2, 2, 1, 3],
                [3, 2, 0, 1, 3, 2],
                [2, 2, 1, 0, 1, 3],
                [2, 1, 3, 1, 0, 1],
                [3, 3, 2, 3, 1, 0],
            ],
            dtype=np.float_,
        )
        labels = np.array([-1, -1, -1, -1, 1, 1], dtype=np.int_)

        count, trees = ortho_mst.ortho_mst_count(weights, labels, 3)
        self.assertEqual(count, 8)
        self.assertEqual(trees, 3)

    def test_orthogonal_mst_3(self):
        """ ORTHO_MST: Making 3 orthogonal MST returns all edges"""
        # We need at least six nodes to make 3 orthogonal MSTs
        weights = np.array(
            [
                [0, 1, 3, 2, 2, 3],
                [1, 0, 2, 2, 1, 3],
                [3, 2, 0, 1, 3, 2],
                [2, 2, 1, 0, 1, 3],
                [2, 1, 3, 1, 0, 1],
                [3, 3, 2, 3, 1, 0],
            ],
            dtype=np.float_,
        )
        labels = np.array([-1, 1, -1, 1, -1, 1], dtype=np.int_)

        count, trees = ortho_mst.ortho_mst_count(weights, labels, 3)
        self.assertEqual(count, 9)
        self.assertEqual(trees, 3)


if __name__ == "__main__":
    unittest.main()



"""

"""

import ctypes
import cython

import numpy as np
cimport numpy as np

cdef extern void c_multiclass_mst_count(double *weights, long *labels, 
        long nTrees, long nObs, long nClass, long *count_ret, long *n_tree_ret)


@cython.boundscheck(False)
@cython.wraparound(False)
def multiclass_mst_count(
        np.ndarray[double, ndim=2, mode="c"] weights not None,
        np.ndarray[long, ndim=1, mode="c"] labels not None,
        long nClass,
        long nTrees):
    """
    """

    cdef long nObs
    cdef long actual_trees

    nObs = weights.shape[0]

    cdef np.ndarray[long, ndim=2, mode='c'] C
    C = np.zeros((nClass, nClass), dtype=ctypes.c_long)

    c_multiclass_mst_count(&weights[0, 0], &labels[0], nTrees, nObs, nClass, 
            &C[0, 0], &actual_trees)

    return C, actual_trees



"""

    Unit tests for the SmartSVM class

"""

from __future__ import division

import math
import networkx as nx
import numpy as np
import six
import unittest

from sklearn.svm import LinearSVC, SVC
from sklearn.model_selection import train_test_split
from sklearn.datasets import load_digits
from smartsvm import SmartSVM

if six.PY2:
    from mock import patch, PropertyMock
else:
    from unittest.mock import patch, PropertyMock


class MockClassifier(object):
    def __init__(self):
        self._X = None
        self._y = None

    def fit(self, X, y):
        self._X = X
        self._y = y
        self.coef_ = None

    def predict(self, X):
        return np.tile(1, (X.shape[0],))


class SmartSVMTestCase(unittest.TestCase):
    def test_init_1(self):
        """ SMARTSVM: Test init defaults """
        inst = SmartSVM()
        self.assertEqual(inst.cut_algorithm, "stoer_wagner")
        self.assertEqual(inst.binary_clf, LinearSVC)
        self.assertEqual(inst.n_jobs, 1)
        self.assertEqual(inst.graph, None)

    def test_init_2(self):
        """ SMARTSVM: Test init supplied """
        inst = SmartSVM(
            cut_algorithm="other", binary_clf="test", n_jobs=4, graph="graph"
        )
        self.assertEqual(inst.cut_algorithm, "other")
        self.assertEqual(inst.binary_clf, "test")
        self.assertEqual(inst.n_jobs, 4)
        self.assertEqual(inst.graph, "graph")

    def test_elements_1(self):
        """ SMARTSVM: Test elements without graph """
        inst = SmartSVM()
        with self.assertRaises(ValueError) as cm:
            x = inst.elements
            del x

        the_exception = cm.exception
        self.assertEqual(
            the_exception.args, ("graph should be a networkx Graph object.",)
        )

    def test_elements_2(self):
        """ SMARTSVM: Test elements with graph """
        G = nx.Graph()
        G.add_edge(1, 2, weight=1.1)
        G.add_edge(1, 3, weight=1.2)
        G.add_edge(1, 4, weight=1.3)
        G.add_edge(2, 3, weight=1.4)
        inst = SmartSVM(graph=G)
        self.assertEqual(sorted(inst.elements), [1, 2, 3, 4])

    @patch.multiple(
        SmartSVM,
        __abstractmethods__=set(),
        is_splitable=PropertyMock(return_value=False),
    )
    def test_fit_1(self):
        """ SMARTSVM: Test fit no graph not splitable """
        inst = SmartSVM(binary_clf=MockClassifier, normalize_error=False)
        # Uses the same data as in test_build_error_graph
        X = np.array(
            [
                [-0.05, 3],
                [0.05, 3],
                [0, 3.00],
                [3, 0.05],
                [3, -0.05],
                [3.00, 0],
                [-0.05, -3],
                [0.05, -3],
                [0, -3.00],
                [-3, 0.05],
                [-3, -0.05],
                [-3.00, 0],
            ]
        )
        y = np.array([1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4])
        exp_weight = (
            1.0 / 2.0
            - 1.0 / 4.0 * math.sqrt(1.0 / 3.0)
            - 1.0 / 4.0 * 1.0 / 3.0
        )
        exp_G = nx.Graph()
        exp_G.add_edge(1, 2, weight=exp_weight)
        exp_G.add_edge(1, 3, weight=exp_weight)
        exp_G.add_edge(1, 4, weight=exp_weight)
        exp_G.add_edge(2, 3, weight=exp_weight)
        exp_G.add_edge(2, 4, weight=exp_weight)
        exp_G.add_edge(3, 4, weight=exp_weight)

        inst.fit(X, y)
        em = nx.algorithms.isomorphism.numerical_edge_match("weight", 1)
        self.assertTrue(nx.is_isomorphic(inst.graph, exp_G, edge_match=em))
        self.assertEqual(sorted(inst.graph.nodes()), sorted(exp_G.nodes()))

    def test_fit_2(self):
        """ SMARTSVM: Test fit no graph splitable """
        inst = SmartSVM(binary_clf=MockClassifier, normalize_error=False)
        X = np.array(
            [
                [-0.05, 3],
                [0.05, 3],
                [0, 3.00],
                [3, 0.05],
                [3, -0.05],
                [3.00, 0],
                [-0.05, -3],
                [0.05, -3],
                [0, -3.00],
            ]
        )
        y = np.array([1, 1, 1, 2, 2, 2, 3, 3, 3])
        inst.fit(X, y)

        inst_y = np.array([-1, -1, -1, -1, -1, -1, 1, 1, 1])
        self.assertTrue(np.array_equal(inst.classifier_._X, X))
        # xor because labelling could be inverted
        self.assertTrue(
            np.array_equal(inst.classifier_._y, inst_y)
            ^ np.array_equal(inst.classifier_._y, -1 * inst_y)
        )

        ninst_X = np.array(
            [
                [-0.05, 3],
                [0.05, 3],
                [0, 3.00],
                [3, 0.05],
                [3, -0.05],
                [3.00, 0],
            ]
        )
        ninst_y = np.array([-1, -1, -1, 1, 1, 1])

        # a bit verbose because of symmetry
        if hasattr(inst.negative_child_, "classifier_"):
            self.assertTrue(
                np.array_equal(inst.negative_child_.classifier_._X, ninst_X)
            )
            self.assertTrue(
                np.array_equal(inst.negative_child_.classifier_._y, ninst_y)
                ^ np.array_equal(
                    inst.negative_child_.classifier_._y, -1 * ninst_y
                )
            )
        else:
            self.assertTrue(
                np.array_equal(inst.positive_child_.classifier_._X, ninst_X)
            )
            self.assertTrue(
                np.array_equal(inst.positive_child_.classifier_._y, ninst_y)
                ^ np.array_equal(
                    inst.positive_child_.classifier_._y, -1 * ninst_y
                )
            )

    def test_fit_predict_proba(self):
        """ SMARTSVM: Test fit and predict_proba """
        X = np.array(
            [
                [-0.05, 3],
                [0.05, 3],
                [0, 3.00],
                [3, 0.05],
                [3, -0.05],
                [3.00, 0],
                [-0.05, -3],
                [0.05, -3],
                [0, -3.00],
                [-3, 0.05],
                [-3, -0.05],
                [-3.00, 0],
            ]
        )
        y = np.array([1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4])
        clf = SmartSVM(binary_clf=SVC, clf_params={"probability": True})
        clf.fit(X, y)
        probs = clf.predict_proba(X)
        # it's easier to cast it as a pandas DataFrame and do idxmax(1).
        preds = [
            max([(v[i], c) for c, v in probs.items()], key=lambda x: x[0])[1]
            for i in range(X.shape[0])
        ]
        self.assertTrue(np.array_equal(preds, y))

    def test_fit_predict_strings(self):
        """ SMARTSVM: Test fit and predict with string targets """
        digits = load_digits(10)
        n_samples = len(digits.images)
        X = digits.images.reshape(n_samples, -1)
        y = digits.target
        yy = np.zeros((len(y),), dtype="a10")
        yy[y == 0] = "red"
        yy[y == 1] = "blue"
        yy[y == 2] = "green"
        yy[y == 3] = "purple"
        yy[y == 4] = "white"
        yy[y == 5] = "grey"
        yy[y == 6] = "black"
        yy[y == 7] = "yellow"
        yy[y == 8] = "brown"
        yy[y == 9] = "pink"

        X_train, X_test, y_train, y_test = train_test_split(
            X, yy, test_size=0.33
        )
        clf = SmartSVM()
        clf.fit(X_train, y_train)
        y_pred = clf.predict(X_test)

        self.assertTrue(np.array_equal(np.unique(y_train), np.unique(y_pred)))

    def test_repr(self):
        """ SMARTSVM: Test repr """
        inst = SmartSVM()
        self.assertEqual(repr(inst), "SmartSVM()")

        inst = SmartSVM()
        inst.negative_ = set([1, 2, 3])
        inst.positive_ = set([4, 5])
        if six.PY2:
            self.assertEqual(
                repr(inst),
                ("SmartSVM(negative=" "set([1, 2, 3]), positive=set([4, 5]))"),
            )
        else:
            self.assertEqual(
                repr(inst),
                ("SmartSVM(negative=" "{1, 2, 3}, positive={4, 5})"),
            )

        inst = SmartSVM()
        inst.graph = nx.Graph()
        inst.graph.add_node(1)
        if six.PY2:
            self.assertEqual(repr(inst), "SmartSVM(set([1]))")
        else:
            self.assertEqual(repr(inst), "SmartSVM({1})")

    def test_split_1(self):
        """ SMARTSVM: test split not splitable """
        graph = nx.Graph()
        graph.add_node(1)
        inst = SmartSVM(graph=graph)
        inst._split()
        self.assertFalse(hasattr(inst, "negative_"))
        self.assertFalse(hasattr(inst, "positive_"))

    def test_split_2(self):
        """ SMARTSVM: test split """
        # Uses the same data as in test_graph_cut_sw
        G = nx.Graph()
        G.add_edge("x", "a", weight=3)
        G.add_edge("x", "b", weight=2)
        G.add_edge("a", "c", weight=3)
        G.add_edge("b", "c", weight=5)
        G.add_edge("b", "d", weight=4)
        G.add_edge("d", "e", weight=2)
        G.add_edge("c", "y", weight=2)
        G.add_edge("e", "y", weight=3)
        G1 = nx.Graph()
        G1.add_edge("x", "a", weight=3)
        G1.add_edge("x", "b", weight=2)
        G1.add_edge("a", "c", weight=3)
        G1.add_edge("b", "c", weight=5)
        G1.add_edge("b", "d", weight=4)
        G2 = nx.Graph()
        G2.add_edge("e", "y", weight=3)

        inst = SmartSVM(graph=G)
        inst._split()

        self.assertTrue(isinstance(inst.negative_child_, SmartSVM))
        self.assertTrue(isinstance(inst.positive_child_, SmartSVM))

        em = nx.algorithms.isomorphism.numerical_edge_match("weight", 1)

        # graphs can be switched
        exp_set_1 = set(["a", "b", "c", "d", "x"])
        exp_set_2 = set(["e", "y"])
        if inst.negative_ == exp_set_1:
            self.assertEqual(inst.negative_, exp_set_1)
            self.assertEqual(inst.positive_, exp_set_2)
            self.assertTrue(
                nx.is_isomorphic(inst.negative_child_.graph, G1, edge_match=em)
            )
            self.assertTrue(
                nx.is_isomorphic(inst.positive_child_.graph, G2, edge_match=em)
            )
        else:
            self.assertEqual(inst.positive_, exp_set_1)
            self.assertEqual(inst.negative_, exp_set_2)
            self.assertTrue(
                nx.is_isomorphic(inst.negative_child_.graph, G2, edge_match=em)
            )
            self.assertTrue(
                nx.is_isomorphic(inst.positive_child_.graph, G1, edge_match=em)
            )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import os
import re
import numpy

# from distutils.core import setup
from setuptools import setup, find_packages
from distutils.extension import Extension

# Set this to True to enable building extensions using Cython. Set it to False
# to build extensions from the C file (that was previously generated using
# Cython). Set it to 'auto' to build with Cython if available, otherwise from
# the C file.
USE_CYTHON = "auto"

if USE_CYTHON:
    try:
        from Cython.Distutils import build_ext
    except ImportError:
        if USE_CYTHON == "auto":
            USE_CYTHON = False
        else:
            raise

cmdclass = {}
ext_modules = []

if USE_CYTHON:
    ext_modules += [
        Extension(
            "smartsvm.ortho_mst",
            [
                os.path.join("src", "ortho_mst.pyx"),
                os.path.join("src", "c_ortho_mst.c"),
            ],
            include_dirs=[numpy.get_include()],
        ),
        Extension(
            "smartsvm.multiclass_mst_count",
            [
                os.path.join("src", "multiclass_mst_count.pyx"),
                os.path.join("src", "c_multiclass_mst_count.c"),
            ],
            include_dirs=[numpy.get_include()],
        ),
    ]
    cmdclass.update({"build_ext": build_ext})
else:
    ext_modules += [
        Extension(
            "smartsvm.ortho_mst",
            [
                os.path.join("src", "ortho_mst.c"),
                os.path.join("src", "c_ortho_mst.c"),
            ],
            include_dirs=[numpy.get_include()],
        ),
        Extension(
            "smartsvm.multiclass_mst_count",
            [
                os.path.join("src", "multiclass_mst_count.c"),
                os.path.join("src", "c_multiclass_mst_count.c"),
            ],
            include_dirs=[numpy.get_include()],
        ),
    ]


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


version = re.search(
    "__version__ = \"([^']+)\"", open("smartsvm/__init__.py").read()
).group(1)

setup(
    name="smartsvm",
    version=version,
    description=(
        "Python package for Meta-Learning and Adaptive "
        "Hierarchical Classifier Design"
    ),
    long_description=read("README.rst"),
    url="https://github.com/HeroResearchGroup/SmartSVM",
    author="G.J.J. van den Burg",
    author_email="gertjanvandenburg@gmail.com",
    license="GPL v2",
    install_requires=["networkx>=1.9", "scipy", "scikit-learn", "numpy"],
    packages=find_packages(),
    cmdclass=cmdclass,
    ext_modules=ext_modules,
    classifiers=[
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
    ],
    zip_safe=False
)



"""
    Unit tests for the cut module of the SmartSVM package.
"""

from __future__ import division

import unittest
import networkx as nx

from smartsvm import cut


class CutTestCase(unittest.TestCase):
    def test_graph_cut_1(self):
        """ CUT: Test main graph_cut function - default args"""
        # This function simply tests if dispatching is done correctly.
        G = nx.Graph()
        G.add_edge("x", "y", weight=3)
        G1_exp = nx.Graph()
        G1_exp.add_node("x")
        G2_exp = nx.Graph()
        G2_exp.add_node("y")

        G1_res, G2_res = cut.graph_cut(G)
        self.assertTrue(
            (sorted(G1_res.nodes()) == sorted(G1_exp.nodes()))
            ^ (sorted(G1_res.nodes()) == sorted(G2_exp.nodes()))
        )
        self.assertTrue(
            (sorted(G2_res.nodes()) == sorted(G1_exp.nodes()))
            ^ (sorted(G2_res.nodes()) == sorted(G2_exp.nodes()))
        )

    def test_graph_cut_2(self):
        """ CUT: Test main graph_cut function - Stoer-Wagner"""
        G = nx.Graph()
        G.add_edge("x", "y", weight=3)
        G1_exp = nx.Graph()
        G1_exp.add_node("x")
        G2_exp = nx.Graph()
        G2_exp.add_node("y")

        G1_res, G2_res = cut.graph_cut(G, algorithm="stoer_wagner")
        self.assertTrue(
            (sorted(G1_res.nodes()) == sorted(G1_exp.nodes()))
            ^ (sorted(G1_res.nodes()) == sorted(G2_exp.nodes()))
        )
        self.assertTrue(
            (sorted(G2_res.nodes()) == sorted(G1_exp.nodes()))
            ^ (sorted(G2_res.nodes()) == sorted(G2_exp.nodes()))
        )

    def test_graph_cut_3(self):
        """ CUT: Test main graph_cut function - Spectral Clustering (1) """
        G = nx.Graph()
        G.add_edge("x", "y", weight=3)
        G.add_edge("x", "z", weight=5)
        G1_exp = nx.Graph()
        G1_exp.add_node("x")
        G1_exp.add_node("y")
        G2_exp = nx.Graph()
        G2_exp.add_node("z")

        G1_res, G2_res = cut.graph_cut(G, algorithm="spectral_clustering")
        self.assertTrue(
            (sorted(G1_res.nodes()) == sorted(G1_exp.nodes()))
            ^ (sorted(G1_res.nodes()) == sorted(G2_exp.nodes()))
        )
        self.assertTrue(
            (sorted(G2_res.nodes()) == sorted(G1_exp.nodes()))
            ^ (sorted(G2_res.nodes()) == sorted(G2_exp.nodes()))
        )

    def test_graph_cut_4(self):
        """ CUT: Test main graph_cut function - Spectral Clustering (2) """
        G = nx.Graph()
        G.add_edge("x", "y", weight=3)
        G.add_edge("x", "z", weight=5)
        G1_exp = nx.Graph()
        G1_exp.add_node("x")
        G1_exp.add_node("y")
        G2_exp = nx.Graph()
        G2_exp.add_node("z")

        G1_res, G2_res = cut.graph_cut(G, algorithm="normalized_cut")
        self.assertTrue(
            (sorted(G1_res.nodes()) == sorted(G1_exp.nodes()))
            ^ (sorted(G1_res.nodes()) == sorted(G2_exp.nodes()))
        )
        self.assertTrue(
            (sorted(G2_res.nodes()) == sorted(G1_exp.nodes()))
            ^ (sorted(G2_res.nodes()) == sorted(G2_exp.nodes()))
        )

    def test_graph_cut_5(self):
        """ CUT: Test main graph_cut function - Unknown algorithm"""
        # This function simply tests if dispatching is done correctly.
        G = nx.Graph()
        G.add_edge("x", "y", weight=3)
        with self.assertRaises(ValueError) as cm:
            cut.graph_cut(G, algorithm="test")

        the_exception = cm.exception
        self.assertEqual(
            the_exception.args, ("Unknown graph cut algorithm supplied: test",)
        )

    def test_graph_cut_sw(self):
        """ CUT: Test Stoer-Wagner graph cut. """
        # This test verifies if the returned graphs are the expected subgraphs.
        # Uses the example of Stoer-Wagner from the networkx docs, slightly
        # modified to make sure the cut is unique.
        G = nx.Graph()
        G.add_edge("x", "a", weight=3)
        G.add_edge("x", "b", weight=2)
        G.add_edge("a", "c", weight=3)
        G.add_edge("b", "c", weight=5)
        G.add_edge("b", "d", weight=4)
        G.add_edge("d", "e", weight=2)
        G.add_edge("c", "y", weight=2)
        G.add_edge("e", "y", weight=3)
        G1_exp = nx.Graph()
        G1_exp.add_edge("x", "a", weight=3)
        G1_exp.add_edge("x", "b", weight=2)
        G1_exp.add_edge("a", "c", weight=3)
        G1_exp.add_edge("b", "c", weight=5)
        G1_exp.add_edge("b", "d", weight=4)
        G2_exp = nx.Graph()
        G2_exp.add_edge("e", "y", weight=3)

        G1_res, G2_res = cut.graph_cut_sw(G)
        em = nx.algorithms.isomorphism.numerical_edge_match("weight", 1)
        self.assertTrue(
            nx.is_isomorphic(G1_res, G1_exp, edge_match=em)
            ^ nx.is_isomorphic(G1_res, G2_exp, edge_match=em)
        )
        self.assertTrue(
            nx.is_isomorphic(G2_res, G1_exp, edge_match=em)
            ^ nx.is_isomorphic(G2_res, G2_exp, edge_match=em)
        )
        self.assertTrue(
            (sorted(G1_res.nodes()) == sorted(G1_exp.nodes()))
            ^ (sorted(G1_res.nodes()) == sorted(G2_exp.nodes()))
        )
        self.assertTrue(
            (sorted(G2_res.nodes()) == sorted(G1_exp.nodes()))
            ^ (sorted(G2_res.nodes()) == sorted(G2_exp.nodes()))
        )


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

"""Functions for computing the Henze-Penrose divergence

This module contains functions for computing the Henze-Penrose divergence 
between data from two classes (:func:`.divergence`) or between one class and 
the collection of all other classes (:func:`.ovr_divergence`).

"""

from __future__ import division

import numpy as np

from math import sqrt
from sklearn.preprocessing import LabelEncoder

from .cross_connections import binary_cross_connections, ovr_cross_connections


def merge_and_label(X1, X2):
    """Merge two datasets

    Merge the datasets into one array and create a vector of labels to preserve 
    class membership.

    Parameters
    ----------
    X1 : numpy.ndarray
        Data matrix for one of the classes

    X2 : numpy.ndarray
        Data matrix for the other class

    Returns
    -------
    data : numpy.ndarray
        Data matrix which vertically stacks X1 and X2

    labels : numpy.ndarray
        Labels vector of -1 and 1 for X1 and X2 data respectively

    """
    # Number of instances in both datasets
    n1 = X1.shape[0]
    n2 = X2.shape[0]

    # Combine data
    data = np.vstack((X1, X2))

    # Preserve class information
    labels1 = np.ones(n1, dtype=np.int_) * -1
    labels2 = np.ones(n2, dtype=np.int_)
    labels = np.hstack((labels1, labels2))

    return data, labels


def divergence(X1, X2, nTrees=3, bias_correction=True):
    """Compute the Henze-Penrose divergence

    Compute the Henze-Penrose divergence between data from two classes using 
    the Friedman-Rafsky statistic. This is based on the Euclidean minimal 
    spanning tree between two classes. To reduce variance in the estimate, a 
    number of orthogonal MSTs are used that can be set with a function 
    parameter. A bias correction to the divergence is applied, unless this is 
    disabled by the user (if disabled, the estimate is still corrected to 
    ensure non-negativity).

    Parameters
    ----------
    X1 : numpy.ndarray
        Data matrix for one of the classes

    X2 : numpy.ndarray
        Data matrix for the other class

    nTrees : int
        Number of orthogonal minimal spanning trees to use

    bias_correction : bool
        Whether or not to apply bias correction to the estimator

    Returns
    -------
    divergence : float
        The Henze-Penrose divergence between the classes

    """

    # Get array dimensions
    n1 = X1.shape[0]
    n2 = X2.shape[0]
    N = n1 + n2

    # Create a merged dataset and keep the labels
    data, labels = merge_and_label(X1, X2)

    # Compute the cross connections
    C = binary_cross_connections(data, labels, nTrees=nTrees)

    # Calculate the divergence measure
    if bias_correction:
        # Calculate the bias-correction factor
        phat = n1 / N
        qhat = n2 / N
        gamma = (
            2.0 * N * min(phat, qhat)
            - 3.0 / 4.0 * N
            + 1.0 / 4.0 * N * sqrt(9.0 - 16.0 * min(phat, qhat))
        )
        Dptilde = 1.0 - 2.0 * min(gamma, C) / N
    else:
        Dptilde = 1.0 - 2.0 * min(N / 2, C) / N

    return Dptilde


def ovr_divergence(X, labels, nTrees=3, bias_correction=True):
    """Compute the One-vs-Rest Henze-Penrose Divergence for all classes

    This function is similar to :func:`.divergence` with the difference that it 
    computes the One-vs-Rest divergence. This is the divergence between one 
    class and the collection of all the other classes together. This divergence 
    is computed for each class simultaneously.

    Parameters
    ----------
    X : numpy.ndarray
        Data matrix

    labels : numpy.ndarray
        Class labels for the instances

    nTrees : int
        The number of orthogonal MSTs to construct

    bias_correction : bool
        Whether or not to apply bias correction

    Returns
    -------
    ovr_divergence : dict
        Dictionary with a mapping from the class label to the OvR divergence estimate

    """

    # Encode the classes
    encoder = LabelEncoder()
    y = encoder.fit_transform(labels)
    nClass = len(encoder.classes_)
    N = X.shape[0]

    Cs = ovr_cross_connections(X, y, nClass, nTrees=nTrees)

    Dptildes = np.zeros((nClass,))
    for i in range(nClass):
        if bias_correction:
            phat = X[y == i, :].shape[0] / N
            qhat = X[y != i, :].shape[0] / N
            gamma = (
                2.0 * N * min(phat, qhat)
                - 3.0 / 4.0 * N
                + 1.0 / 4.0 * N * sqrt(9.0 - 16.0 * min(phat, qhat))
            )
            Dptildes[i] = 1.0 - 2.0 * min(gamma, Cs[i]) / N
        else:
            Dptildes[i] = 1.0 - 2.0 * min(N / 2, Cs[i]) / N

    ovr_divs = {}
    for i, dp in zip(encoder.classes_, Dptildes):
        ovr_divs[i] = dp
    return ovr_divs
